import asyncio
import socket
import numpy as np
import math
import pandas as pd
import time
import json
from collections import defaultdict
from itertools import permutations, product, combinations
import argparse
import os
import pkg_resources


class Server:
    def __init__(self, config_file, ip=None, port=None):
        
        config_path = pkg_resources.resource_filename('agt_server', config_file)
        with open(config_path) as cfile:
            server_config = json.load(cfile)
        self.n_players = 0  
        
        self.logging_path = pkg_resources.resource_filename('agt_server', f"server/logging/temp_res_table.txt")
        self.agent_path = pkg_resources.resource_filename('agt_server', f"server/logging/temp_agent_data.txt")

        os.makedirs(os.path.dirname(self.logging_path), exist_ok=True)
        os.makedirs(os.path.dirname(self.agent_path), exist_ok=True)

        if port != None:
            self.port = port
        else:
            self.port = 8080

        if ip != None:
            self.ip = ip
        else:
            hostname = socket.gethostname()
            self.ip = socket.gethostbyname(hostname)

        self.server_config = server_config
        self.game_name = server_config['game_name']
        self.signup_time = server_config['signup_time']
        self.player_types = server_config['player_types']
        self.permission_map = server_config['permissions']
        self.type_configurations = server_config['type_configurations']
        self.players_per_game = server_config['num_players_per_game']
        self.num_rounds = server_config['num_rounds'] + 1
        self.round_buffer = server_config['round_buffer']
        self.save_results = server_config['save_results']
        self.save_path = server_config['save_path']
        self.display_results = server_config['display_results']
        self.kick_time = server_config['kick_time']
        self.send_results = server_config['send_results']
        self.check_dev_id = server_config['check_dev_id']
        self.invalid_move_penalty = server_config['invalid_move_penalty']
        self.timeout_tolerance = server_config['timeout_tolerance']
        self.df = None

        game_path = server_config['game_path']
        module_name, class_name = game_path.rsplit('.', 1)
        game_module = __import__(module_name, fromlist=[class_name])
        self.game = getattr(game_module, class_name)
        self.order_matters = self.game.order_matters

        self.curr_round = 1
        self.player_data = defaultdict(lambda:
                                       {
                                           "name": None,
                                           "client": None,
                                           "device_id": None,
                                           "index": None, 
                                           "ingame": False, 
                                           "disconnected": False
                                       })
        self.result_table = None
        self.can_run_game = asyncio.Event()
        
    async def countdown(self):
        while self.signup_time > 0:
            print(f"{self.signup_time} seconds remaining")
            await asyncio.sleep(1)
            self.signup_time -= 1
        
        if self.players_per_game == 2: 
            self.result_table = np.zeros([self.n_players, self.n_players])
        else: 
            self.result_table = []

        message = {"message": "provide_game_name",
                   "game_name": self.game_name}
        for address in self.player_data:
            writer, reader = self.player_data[address]['client']
            try: 
                writer.write(json.dumps(message).encode())
                await writer.drain()
                data = await reader.read(1024)
                response = json.loads(data)
                assert response['message'] == "game_name_recieved"
            except Exception as e: 
                print(f"Failed to send game_name to {self.player_data[address]['name']}... Most likely disconnected")
                writer.close()
                await writer.wait_closed()
                self.player_data[address]['client'] = None
                self.player_data[address]['device_id'] = None
                self.player_data[address]['address'] = None
                self.player_data[address]['disconnected'] = True
                continue
        self.can_run_game.set() 
        
        return
    
    
    async def handle_client(self, reader, writer):
        address = writer.get_extra_info('peername')
        
        if self.signup_time <= 0:
            print(f"Client {address} has failed to connect [Sign Up Expired]")
            try: 
                writer.close()
                await asyncio.wait_for(writer.wait_closed(), timeout=0.5)
            except asyncio.TimeoutError:
                pass
            return
        
        message = {"message": "request_device_id"}
        try: 
            writer.write(json.dumps(message).encode())
            await writer.drain()
            data = await reader.read(1024)
        except Exception as e: 
            print(f"Client {address} has failed to connect because {type(e).__name__}: {e}")
            try: 
                writer.close()
                await asyncio.wait_for(writer.wait_closed(), timeout=0.5)
            except asyncio.TimeoutError:
                pass
            return
        
        # No response from client
        if not data:
            print(f"Client {address} has failed to connect [No Response]")
            writer.close()
            await writer.wait_closed()
            return
        try:
            response = json.loads(data)
        except:
            print(f"Client {address} has failed to connect [Invalid Response]")
            writer.close()
            await writer.wait_closed()
            return
        
        if response['message'] != "provide_device_id" or response['device_id'] == None:
            print(f"Client {address} has failed to connect [No Device ID Provided]")
            writer.close()
            await writer.wait_closed()
            return
        
        device_id = response['device_id']
        
        self.n_players += 1
        print(f"Accepted connection from {address} ({device_id})")
        
        # Check if device is already connected
        if self.check_dev_id:
            if device_id in [pd['device_id'] for pd in self.player_data.values()]:
                print(f"Client {address} ({device_id}) is already connected")
                writer.close()
                await writer.wait_closed()
                self.n_players -= 1
                return
        
        message = {"message": "request_name"}
        try: 
            writer.write(json.dumps(message).encode())
            await writer.drain()
            data = await reader.read(1024)
        except Exception as e: 
            print(f"Failed to send recieve name from {address} because {type(e).__name__}: {e}")
            writer.close()
            await writer.wait_closed()

        # No response from client
        if not data:
            print(f"Client {address} has disconnected unexpectedly")
            writer.close()
            await writer.wait_closed()
            self.n_players -= 1
            return
    
        try:
            response = json.loads(data)
        except:
            print(f"Client {address} has disconnected unexpectedly")
            print("Please check if their name is too long, the name can be at most 900 characters long")
            writer.close()
            await writer.wait_closed()
            self.n_players -= 1
            return

        if response['message'] != "provide_name" or response['name'] == None:
            print(f"Client {address} did not provide name")
            writer.close()
            await writer.wait_closed()
            self.n_players -= 1
            return

        # Add player to active list
        self.player_data[address]['client'] = (writer, reader)
        self.player_data[address]['device_id'] = device_id
        self.player_data[address]['index'] = self.n_players - 1
        self.player_data[address]['address'] = address

        response['name'] = response['name'].strip().replace("\t", "").replace("\n", "")
        counter = 0
        extension = ""
        while any([response['name'] + extension == data['name'] for data in self.player_data.values()]):
            extension = f" ({counter + 1})"
            counter += 1
        self.player_data[address]['name'] = response['name'] + extension
        message = {"message": "provide_name", "name": self.player_data[address]['name']}
        try: 
            writer.write(json.dumps(message).encode())
            await writer.drain()
            data = await reader.read(1024)
            response = json.loads(data)
            assert response['message'] == "name_updated"
        except Exception as e: 
            print(f"Client {address} has failed to connect because {type(e).__name__}: {e}")
            writer.close()
            await writer.wait_closed()
            self.player_data[address]['client'] = None
            self.player_data[address]['device_id'] = None
            self.player_data[address]['address'] = None
            self.player_data[address]['disconnected'] = True
            self.n_players -= 1
            return
        
        print(f"{self.player_data[address]['name']} has joined the game \n")
    
    async def run_server(self, host, port):
        self.signup_start_time = time.time()
        countdown_task = asyncio.create_task(self.countdown())
        gamerunning_task = asyncio.create_task(self.run_game())
        server = await asyncio.start_server(self.handle_client, host, port)

        async with server:
            try:
                await gamerunning_task
                await self.close()
            except asyncio.CancelledError:
                if not gamerunning_task.done():
                    gamerunning_task.cancel()
                await gamerunning_task
            finally:
                server.close()
                await server.wait_closed()
                if not countdown_task.done():
                    countdown_task.cancel()
                await asyncio.gather(countdown_task, return_exceptions=True)
                
            
    async def start(self):
        try:
            print(f'The server is hosted at {self.ip} and port {self.port}')
            await self.run_server(self.ip, self.port)
        except KeyboardInterrupt:
            print("KeyboardInterrupt caught, finalizing game...")
        return
    
    async def process_game_results(self, game_task, addresses):
        try:
            game_reports = await game_task
        except Exception as e:
            print(f"Error in game with {self.player_data[address]['name']} because {type(e).__name__}: {e}")
            for address in addresses:
                self.player_data[address]['disconnected'] = True
            return

        for address in game_reports:
            if game_reports[address]['disconnected']:
                print(f"Client {self.player_data[address]['name']}: {address} has disconnected unexpectedly")
                self.player_data[address]['client'] = None
                self.player_data[address]['device_id'] = None
                self.player_data[address]['address'] = None
                self.player_data[address]['disconnected'] = True
                self.n_players -= 1
            else:
                total_util = sum(game_reports[address]['util_history'])
                if self.players_per_game == 2: 
                    for opp_address in game_reports:
                        if address != opp_address:
                            self.result_table[self.player_data[address]['index'],
                                            self.player_data[opp_address]['index']] += total_util
                            np.savetxt(self.logging_path, self.result_table, fmt='%d')
                            data_store = [{
                                            "name": self.player_data[key]['name'],
                                            "client": None if self.player_data[key]['client'] == None else "Exists"
                                          } for key in self.player_data]
                            
                            try:
                                with open(self.agent_path, 'w') as json_file:
                                    json.dump(data_store, json_file)
                            except Exception as e:
                                print(f"An error occurred: {e}")

                            
            if self.players_per_game > 2: 
                adds = []
                total_utils = []
                winner, ws = None, float("-inf")
                for address in game_reports: 
                    adds.append(address)
                    total_util = sum(game_reports[address]['util_history'])
                    total_utils.append(total_util)
                    if total_util > ws: 
                        winner = self.player_data[address]['name']
                        ws = total_util
                self.result_table.append(adds + total_utils + [winner])

                with open(self.logging_path, 'w') as file:
                    json.dump(self.result_table, file)
                
                data_store = [{
                                "name": self.player_data[key]['name'],
                                "client": None if self.player_data[key]['client'] == None else "Exists"
                                } for key in self.player_data]
                
                try:
                    with open(self.agent_path, 'w') as json_file:
                        json.dump(data_store, json_file)
                except Exception as e:
                    print(f"An error occurred: {e}")

            elif self.players_per_game == -1:
                adds = []
                total_utils = []
                winner, ws = None, float("-inf")
                for address in game_reports:
                    adds.append(address)
                    total_util = sum(game_reports[address]['util_history'])
                    total_utils.append(total_util)
                    if total_util > ws:
                        winner = self.player_data[address]['name']
                        ws = total_util
                # Append a row with agent keys, their corresponding total utilities, and the winner
                self.result_table.append(adds + total_utils + [winner])
                
                with open(self.logging_path, 'w') as file:
                    json.dump(self.result_table, file)
                
                data_store = [{
                    "name": self.player_data[key]['name'],
                    "client": None if self.player_data[key]['client'] is None else "Exists"
                } for key in self.player_data]
                
                try:
                    with open(self.agent_path, 'w') as json_file:
                        json.dump(data_store, json_file)
                except Exception as e:
                    print(f"An error occurred: {e}")

        for address in addresses: 
            self.player_data[address]["ingame"] = False

    async def run_game(self):
        await self.can_run_game.wait()
        print(f'I have {self.n_players} agents connected and I am starting the {self.game_name} game')
        
        game_tasks = []
        if self.players_per_game == -1:
            addresses = list(self.player_data.keys())
            if any(self.player_data[address]["disconnected"] for address in addresses):
                pass
            elif all(not self.player_data[address]["ingame"] for address in addresses):
                for address in addresses:
                    self.player_data[address]["ingame"] = True
                if self.type_configurations == "all":
                    type_configs = product(self.player_types, repeat=len(addresses))
                else:
                    type_configs = self.type_configurations
                for player_types in type_configs:
                    game_player_data = [self.player_data[address] for address in addresses]
                    if any(pd['client'] is None for pd in game_player_data):
                        continue
                    game = self.game(
                        self.server_config, 
                        num_rounds = self.num_rounds, 
                        player_data = game_player_data,
                        player_types = player_types,
                        permissions_map = self.permission_map,
                        game_kick_timeout = self.kick_time,
                        game_name = self.game_name,
                        invalid_move_penalty = self.invalid_move_penalty,
                        timeout_tolerance = self.timeout_tolerance
                    )
                    game_task = asyncio.create_task(game.run_game())
                    result_processing_task = asyncio.create_task(self.process_game_results(game_task, addresses))
                    game_tasks.append(result_processing_task)
            await asyncio.gather(*game_tasks, return_exceptions=True)
        else:
            if self.order_matters:
                pairings = list(permutations(self.player_data, r=self.players_per_game))
            else:
                pairings = list(combinations(self.player_data, r=self.players_per_game))
            while pairings: 
                new_pairings = []
                for addresses in pairings: 
                    if any(self.player_data[address]["disconnected"] for address in addresses):
                        continue
                    elif all(not self.player_data[address]["ingame"] for address in addresses): 
                        for address in addresses: 
                            self.player_data[address]["ingame"] = True 
                        if self.type_configurations == "all":
                            type_configs = product(self.player_types, repeat=self.players_per_game)
                        else:
                            type_configs = self.type_configurations
                        for player_types in type_configs:
                            game_player_data = [self.player_data[address] for address in addresses]
                            if any(pd['client'] is None for pd in game_player_data):
                                continue
                            game = self.game(
                                self.server_config, 
                                num_rounds = self.num_rounds,
                                player_data = game_player_data,
                                player_types = player_types,
                                permissions_map = self.permission_map,
                                game_kick_timeout = self.kick_time,
                                game_name = self.game_name,
                                invalid_move_penalty = self.invalid_move_penalty,
                                timeout_tolerance = self.timeout_tolerance
                            )
                            game_task = asyncio.create_task(game.run_game())
                            result_processing_task = asyncio.create_task(self.process_game_results(game_task, addresses))
                            game_tasks.append(result_processing_task)
                    else: 
                        new_pairings.append(addresses)
                            
                pairings = new_pairings
                await asyncio.gather(*game_tasks, return_exceptions=True)

                
        if self.players_per_game == 2: 
            df = pd.DataFrame(self.result_table)
            agent_names = [pld['name'] for pld in self.player_data.values()]
            df.columns = agent_names
            df.index = agent_names
            means = []
            for i, pld in enumerate(self.player_data.values()):
                if pld['client'] == None:
                    self.result_table[i, :] = 0
                    self.result_table[:, i] = 0 
            
            for pld, d in zip(self.player_data.values(), self.result_table):
                if pld['client'] == None:
                    means.append(float('-inf'))
                elif self.n_players <= 1:
                    means.append(0)
                else:
                    means.append(sum(d) / (self.n_players - 1))

            df['Mean Points'] = means
            final_scores = [m / (self.num_rounds * math.factorial(self.players_per_game)) for m in means]
            df['Final Score'] = final_scores
            df = df.sort_values('Mean Points', ascending=False)
            df['Mean Points'] = np.where(df['Mean Points'] == float(
                '-inf'), 'Disconnected', df['Mean Points'])
            df['Final Score'] = np.where(df['Final Score'] == float(
                '-inf'), 'Disconnected', df['Final Score'])
        elif self.players_per_game == 3: 
            extended_results = [
                [self.player_data[a]['name'], self.player_data[b]['name'], self.player_data[c]['name'], *rest] for 
                a, b, c, *rest in self.result_table
            ]
            df = pd.DataFrame(extended_results)
            df.columns = ['Agent 1', 'Agent 2', 'Agent 3',
                        'Agent 1 Score', 'Agent 2 Score', 'Agent 3 Score', 'Winner']
            print(f"Extended Results: \n {df}")

            total_util_dict = defaultdict(lambda: [0, 0])
            for p1, p2, p3, p1_util, p2_util, p3_util, _ in self.result_table:
                total_util_dict[p1][0] += p1_util
                total_util_dict[p2][0] += p2_util
                total_util_dict[p3][0] += p3_util
                total_util_dict[p1][1] += 1
                total_util_dict[p2][1] += 1
                total_util_dict[p3][1] += 1

            res_summary = []
            for key, value in total_util_dict.items():
                name = self.player_data[key]['name']
                if self.player_data[key]['client'] == None:
                    res_summary.append([name, float('-inf'), float('-inf')])
                elif value[1] > 0:
                    res_summary.append(
                        [name, value[0] / value[1], value[0] / (value[1] * self.num_rounds)])
                else:
                    res_summary.append([name, 0, 0])

            df = pd.DataFrame(res_summary)
            df.columns = ['Agent Name', 'Average Utility', 'Final Score']
            df = df.sort_values('Final Score', ascending=False)
            df['Average Utility'] = df['Average Utility'].replace(float('-inf'), 'Disconnected')
            df['Final Score'] = df['Final Score'].replace(float('-inf'), 'Disconnected')
            
        elif self.players_per_game == 4: 
            extended_results = [
                [self.player_data[a]['name'], self.player_data[b]['name'], self.player_data[c]['name'], self.player_data[d]['name'], *rest] for
                a, b, c, d, *rest in self.result_table
            ]
            df = pd.DataFrame(extended_results)
            df.columns = ['Agent 1', 'Agent 2', 'Agent 3', 'Agent 4',
                        'Agent 1 Score', 'Agent 2 Score', 'Agent 3 Score', 'Agent 4 Score', 'Winner']
            print(f"Extended Results: \n {df}")

            total_util_dict = defaultdict(lambda: [0, 0])
            for p1, p2, p3, p4, p1_util, p2_util, p3_util, p4_util, _ in self.result_table:
                total_util_dict[p1][0] += p1_util
                total_util_dict[p2][0] += p2_util
                total_util_dict[p3][0] += p3_util
                total_util_dict[p4][0] += p4_util
                total_util_dict[p1][1] += 1
                total_util_dict[p2][1] += 1
                total_util_dict[p3][1] += 1
                total_util_dict[p4][1] += 1

            res_summary = []
            for key, value in total_util_dict.items():
                name = self.player_data[key]['name']
                if self.player_data[key]['client'] == None:
                    res_summary.append([name, float('-inf'), float('-inf')])
                elif value[1] > 0:
                    res_summary.append(
                        [name, value[0] / value[1], value[0] / (value[1] * self.num_rounds)])
                else:
                    res_summary.append([name, 0, 0])

            df = pd.DataFrame(res_summary)
            df.columns = ['Agent Name', 'Average Utility', 'Final Score']
            df = df.sort_values('Final Score', ascending=False)
            df['Average Utility'] = df['Average Utility'].replace(float('-inf'), 'Disconnected')
            df['Final Score'] = df['Final Score'].replace(float('-inf'), 'Disconnected')
            
        elif self.players_per_game == -1:
            # All players participated in the same game.
            num_agents = len(self.player_data)
            # Initialize a dictionary to hold total utility and game count for each agent.
            total_util_dict = defaultdict(lambda: [0, 0])
            
            # Assume each row in result_table has structure:
            # [agent_key_1, agent_key_2, ..., agent_key_N, score_1, score_2, ..., score_N, (optional winner)]
            for row in self.result_table:
                # First num_agents entries are agent keys; next num_agents entries are their scores.
                for i in range(num_agents):
                    agent_key = row[i]
                    score = row[num_agents + i]
                    total_util_dict[agent_key][0] += score
                    total_util_dict[agent_key][1] += 1

            # Build a summary list: for each agent, compute average utility and a final score.
            res_summary = []
            for key, (total, count) in total_util_dict.items():
                name = self.player_data[key]['name']
                if self.player_data[key]['client'] is None:
                    res_summary.append([name, float('-inf'), float('-inf')])
                elif count > 0:
                    avg_util = total / count
                    # Here, final score is calculated by normalizing average utility.
                    final_score = avg_util / self.num_rounds
                    res_summary.append([name, avg_util, final_score])
                else:
                    res_summary.append([name, 0, 0])
            
            df = pd.DataFrame(res_summary, columns=['Agent Name', 'Average Utility', 'Final Score'])
            df = df.sort_values('Final Score', ascending=False)
            # Replace any -inf values with 'Disconnected'
            df['Average Utility'] = df['Average Utility'].replace(float('-inf'), 'Disconnected')
            df['Final Score'] = df['Final Score'].replace(float('-inf'), 'Disconnected')
        
        if self.save_results:
            timestamp = time.strftime("%b %d %Y %H:%M:%S")
            df.to_csv(f"{self.save_path}/data.csv")
        self.df = df
        if self.display_results:
            print(df)
        return df

    async def close(self):
        message = {"message": "game_end",
                   "send_results": self.send_results,
                   "results": self.df.to_json()}

        try:
            os.remove(logging_path)
            os.remove(agent_path)
        except Exception as e: 
            pass
        
        if not self.send_results:
            message['results'] = "Please check in with your Lab TA for the results"

        for address in self.player_data:
            if self.player_data[address]['client']:
                writer, _ = self.player_data[address]['client']
                try: 
                    writer.write(json.dumps(message).encode())
                    await writer.drain()
                except Exception as e: 
                    print(f"Failed to send final results to {self.player_data[address]['name']} because {type(e).__name__}: {e}")
                    continue



async def main(full_config, ip, port):
    server = Server(full_config, ip, port)
    await server.start()

if __name__ == "__main__":
    try: 
        parser = argparse.ArgumentParser(description='My Server')
        parser.add_argument('server_config', type=str,
                            help='Relative Path of the server config file starting from config/server_configs')
        parser.add_argument('--ip', type=str, help='IP address to bind the server to')
        parser.add_argument('--port', type=int, help='Port number to bind the server to')

        args = parser.parse_args()
        full_config = f"configs/server_configs/{args.server_config}"
        
        start_time = time.time()
        asyncio.run(main(full_config, args.ip, args.port))
        end_time = time.time()
        print(f"Server ran in {end_time - start_time} seconds")
        try:
            logging_path = pkg_resources.resource_filename('agt_server', f"server/logging/temp_res_table.txt")
            agent_path = pkg_resources.resource_filename('agt_server', f"server/logging/temp_agent_data.txt")
            os.remove(logging_path)
            os.remove(agent_path)
        except Exception as e: 
            pass
    except KeyboardInterrupt:        
        config_path = pkg_resources.resource_filename('agt_server', full_config)
        with open(config_path) as cfile:
            server_config = json.load(cfile)
                    
        num_players = server_config['num_players_per_game']
        
        logging_path = pkg_resources.resource_filename('agt_server', f"server/logging/temp_res_table.txt")
        agent_path = pkg_resources.resource_filename('agt_server', f"server/logging/temp_agent_data.txt")
        
        with open(agent_path, 'r') as json_file:
            player_data = json.load(json_file)
        
        if num_players == 2: 
            result_table = np.loadtxt(logging_path, dtype=int)
            
            df = pd.DataFrame(result_table)
            agent_names = [pld['name'] for pld in player_data]
            df.columns = agent_names
            df.index = agent_names
            means = []
            for i, pld in enumerate(player_data):
                if pld['client'] == None:
                    result_table[i, :] = 0
                    result_table[:, i] = 0 
            
            for pld, d in zip(player_data, result_table):
                if pld['client'] == None:
                    means.append(float('-inf'))
                else:
                    means.append(sum(d))

            df['Total Points'] = means
            df = df.sort_values('Total Points', ascending=False)
            df['Total Points'] = df['Total Points'].replace(float('-inf'), 'Disconnected')
            print(df)
        # TODO: Refactor to work with more than 2 players. 
        try:
            os.remove(logging_path)
            os.remove(agent_path)
        except Exception as e: 
            pass
        
        




from q_learning import QLearning
from agt_server.local_games.chicken_arena import ChickenArena
from agt_server.agents.test_agents.chicken.basic_agent.my_agent import BasicAgent
import argparse

NUM_TRAINING_ITERATIONS = 20000
NUM_ITERATIONS_PER_PRINT = 1000


class LastMoveChicken(QLearning):
    def __init__(self, name, num_possible_states, num_possible_actions, initial_state, learning_rate, discount_factor, exploration_rate, training_mode, save_path=None) -> None:
        super().__init__(name, num_possible_states, num_possible_actions, initial_state,
                         learning_rate, discount_factor, exploration_rate, training_mode, save_path)

    def determine_state(self):
        return self.get_opp_last_action()


NUM_TRAINING_ITERATIONS = 20000
NUM_ITERATIONS_PER_PRINT = 1000
# this agent only uses 2 states -- one per opponent's possible last action
NUM_POSSIBLE_STATES = 2
# chicken has 2 possible actions (CONTINUE and SWERVE).
NUM_POSSIBLE_ACTIONS = 2
INITIAL_STATE = 0

LEARNING_RATE = 0.05
DISCOUNT_FACTOR = 0.90
EXPLORATION_RATE = 0.05

################### SUBMISSION #####################
agent_submission = LastMoveChicken("LastMove", NUM_POSSIBLE_STATES, NUM_POSSIBLE_ACTIONS,
                                   INITIAL_STATE, LEARNING_RATE, DISCOUNT_FACTOR, EXPLORATION_RATE, True, None)
####################################################

if __name__ == "__main__":

    #### DO NOT TOUCH THIS #####
    parser = argparse.ArgumentParser(description='My Agent')
    parser.add_argument('agent_name', type=str, help='Name of the agent')
    parser.add_argument('--join_server', action='store_true',
                        help='Connects the agent to the server')
    parser.add_argument('--ip', type=str, default='127.0.0.1',
                        help='IP address (default: 127.0.0.1)')
    parser.add_argument('--port', type=int, default=8080,
                        help='Port number (default: 8080)')
    parser.add_argument('--train', type=bool, default=True,
                        help='Train the qtable (default: True)')
    parser.add_argument('--save_path', type=str, default=None,
                        help='Path to save the qtable (default: None)')
    args = parser.parse_args()

    # START SIMULATING THE GAME
    agent = LastMoveChicken(args.agent_name, NUM_POSSIBLE_STATES, NUM_POSSIBLE_ACTIONS,
                            INITIAL_STATE, LEARNING_RATE, DISCOUNT_FACTOR, EXPLORATION_RATE, args.train, args.save_path)
    if args.join_server:
        agent.connect(ip=args.ip, port=args.port)
    else:
        arena = ChickenArena(
            num_rounds=1000,
            timeout=1,
            players=[
                agent,
                BasicAgent("Agent_1"),
                BasicAgent("Agent_2"),
                BasicAgent("Agent_3"),
                BasicAgent("Agent_4")
            ]
        )
        arena.run()



from collections import defaultdict
class GameReport(): 
    def __init__(self) -> None:
        self.game_history = defaultdict(lambda: [])
        
    def __str__(self) -> str:
        return f"Game Report: \n {self.game_history}"
    
    def __repr__(self) -> str:
        return self.__str__()

    def get_game_report_as_dict(self): 
        """
        Retrieves the entire game history as a dictionary.

        :return: The game_history attribute containing all recorded game data.
        """
        return self.game_history
    
    def get_action_history(self):
        """
        Retrieves the history of actions performed by the current player.

        :return: A list of the current player's actions if available; otherwise, an empty list.
        """
        if 'my_action_history' in self.game_history:
            return self.game_history['my_action_history']
        else: 
            return []

    def get_util_history(self):
        """
        Retrieves the utility history for the current player.

        :return: A list of utility values for the current player if available; otherwise, an empty list.
        """
        if 'my_utils_history' in self.game_history:
            return self.game_history['my_utils_history']
        else: 
            return []


    def get_opp_action_history(self):
        """
        Retrieves the action history for the opponent player.

        :return: A list of the opponent player's actions if available; otherwise, an empty list.
        """
        if 'opp_action_history' in self.game_history:
            return self.game_history['opp_action_history']
        else: 
            return []

    def get_opp1_action_history(self): 
        """
        Retrieves the action history for the first opponent player in games with multiple opponents.

        :return: A list of the first opponent player's actions if available; otherwise, an empty list.
        """
        if "opp1_action_history" in self.game_history:
            return self.game_history['opp1_action_history']
        else: 
            return []

    
    def get_opp2_action_history(self): 
        """
        Retrieves the action history for the second opponent player in games with multiple opponents.

        :return: A list of the second opponent player's actions if available; otherwise, an empty list.
        """
        if "opp2_action_history" in self.game_history:
            return self.game_history['opp2_action_history']
        else: 
            return []

    def get_opp_util_history(self):
        """
        Retrieves the utility history for the opponent player.

        :return: A list of utility values for the opponent player if available; otherwise, an empty list.
        """
        if 'opp_utils_history' in self.game_history:
            return self.game_history['opp_utils_history']
        else: 
            return []
    
    def get_opp1_util_history(self):
        """
        Retrieves the utility history for the first opponent player in games with multiple opponents.

        :return: A list of utility values for the first opponent player if available; otherwise, an empty list.
        """ 
        if "opp1_utils_history" in self.game_history:
            return self.game_history['opp1_utils_history']
        else: 
            return []
    
    def get_opp2_util_history(self): 
        """
        Retrieves the utility history for the second opponent player in games with multiple opponents.

        :return: A list of utility values for the second opponent player if available; otherwise, an empty list.
        """
        if "opp2_utils_history" in self.game_history:
            return self.game_history['opp2_utils_history']
        else: 
            return []
    
    def get_mood_history(self):
        """
        Retrieves the mood history for the game, reflecting the emotional state or mood changes over time.

        :return: A list of mood states throughout the game if available; otherwise, an empty list.
        """
        if 'mood_history' in self.game_history:
            return self.game_history['mood_history']
        else: 
            return []
    
    def get_bid_history(self): 
        """
        Retrieves the history of bids made during the game [AS A NDARRAY].

        :return: A list of bids (np.ndarrays) if available; otherwise, an empty list.
        """
        if 'bid_history' in self.game_history:
            return self.game_history['bid_history']
        else: 
            return []
    
    def get_bid_history_map(self): 
        """
        Retrieves the history of bids made during the game [AS A MAP].

        :return: A list of bid (maps from goods to value) if available; otherwise, an empty list.
        """
        if 'bid_history_map' in self.game_history:
            return self.game_history['bid_history_map']
        else: 
            return []
        
    def get_price_history(self): 
        """
        Retrieves the history of prices during the game [AS A NDARRAY].

        :return: A list of prices (np.ndarrays) if available; otherwise, an empty list.
        """
        if 'price_history' in self.game_history:
            return self.game_history['price_history']
        else: 
            return []
    
    def get_price_history_map(self): 
        """
        Retrieves the history of prices during the game [AS A MAP].

        :return: A list of prices (maps from goods to value) if available; otherwise, an empty list.
        """
        if 'price_history_map' in self.game_history:
            return self.game_history['price_history_map']
        else: 
            return []
    
    def get_winner_history(self): 
        """
        Retrieves the history of tentative winners during the game [AS A NDARRAY of OBJECTS].

        :return: A list of winners (np.ndarrays) if available; otherwise, an empty list.
        """
        if 'winner_history' in self.game_history:
            return self.game_history['winner_history']
        else: 
            return []
    
    def get_winner_history_map(self): 
        """
        Retrieves the history of tentative winners during the game [AS A MAP].

        :return: A list of winners (maps from goods to winner) if available; otherwise, an empty list.
        """
        if 'winner_history_map' in self.game_history:
            return self.game_history['winner_history_map']
        else: 
            return []

    def get_last_action(self):
        """
        Retrieves the last action performed by the current player.

        :return: The most recent action of the current player if available; otherwise, None.
        """
        if 'my_action_history' in self.game_history and len(self.game_history['my_action_history']) > 0:
            return self.game_history['my_action_history'][-1]
        
    def get_previous_util(self):
        """
        Retrieves the last utility value for the current player.

        :return: The most recent utility value of the current player if available; otherwise, None.
        """
        if 'my_utils_history' in self.game_history and len(self.game_history['my_utils_history']) > 0:
            return self.game_history['my_utils_history'][-1]

    def get_opp_last_action(self):
        """
        Retrieves the last action performed by the opponent player.

        :return: The most recent action of the opponent player if available; otherwise, None.
        """
        if 'opp_action_history' in self.game_history and len(self.game_history['opp_action_history']) > 0:
            return self.game_history['opp_action_history'][-1]

    def get_opp1_last_action(self):
        """
        Retrieves the last action performed by the first opponent player in games with multiple opponents.

        :return: The most recent action of the first opponent player if available; otherwise, None.
        """
        if 'opp1_action_history' in self.game_history and len(self.game_history['opp1_action_history']) > 0:
            return self.game_history['opp1_action_history'][-1]
    
    def get_opp2_last_action(self):
        """
        Retrieves the last action performed by the second opponent player in games with multiple opponents.

        :return: The most recent action of the second opponent player if available; otherwise, None.
        """
        if 'opp2_action_history' in self.game_history and len(self.game_history['opp2_action_history']) > 0:
            return self.game_history['opp2_action_history'][-1]
    
    def get_opp_last_util(self):
        """
        Retrieves the last utility value for the opponent player.

        :return: The most recent utility value of the opponent player if available; otherwise, None.
        """
        if 'opp_utils_history' in self.game_history and len(self.game_history['opp_utils_history']) > 0:
            return self.game_history['opp_utils_history'][-1]
    
    def get_opp1_last_util(self):
        """
        Retrieves the last utility value for the first opponent player in games with multiple opponents.

        :return: The most recent utility value of the first opponent player if available; otherwise, None.
        """
        if 'opp1_utils_history' in self.game_history and len(self.game_history['opp1_utils_history']) > 0:
            return self.game_history['opp1_utils_history'][-1]
    
    def get_opp2_last_util(self):
        """
        Retrieves the last utility value for the second opponent player in games with multiple opponents.

        :return: The most recent utility value of the second opponent player if available; otherwise, None.
        """
        if 'opp2_utils_history' in self.game_history and len(self.game_history['opp2_utils_history']) > 0:
            return self.game_history['opp2_utils_history'][-1]
    
    def get_last_mood(self):
        """
        Retrieves the most recent mood state from the mood history.

        :return: The latest mood state if available; otherwise, None.
        """
        if 'mood_history' in self.game_history and len(self.game_history['mood_history']) > 0:
            return self.game_history['mood_history'][-1]

    def get_previous_bid(self):
        """
        Retrieves the most recent entry from the bid history as a ndarray. 
        :return: The latest bids if available; otherwise, None.
        """
        if 'bid_history' in self.game_history and len(self.game_history['bid_history']) > 0:
            return self.game_history['bid_history'][-1]
    
    def get_previous_bid_map(self):
        """
        Retrieves the most recent entry from the bid history as a map. 
        :return: The latest bids if available; otherwise, None.
        """
        if 'bid_history_map' in self.game_history and len(self.game_history['bid_history_map']) > 0:
            return self.game_history['bid_history_map'][-1]
    
    def get_previous_price(self): 
        """
        Retrieves the most recent entry from the price history as a ndarray. 
        :return: The latest price if available; otherwise, None.
        """
        if 'price_history' in self.game_history and len(self.game_history['price_history']) > 0:
            return self.game_history['price_history'][-1]
    
    def get_previous_price_map(self): 
        """
        Retrieves the most recent entry from the price history as a map. 
        :return: The latest price if available; otherwise, None.
        """
        if 'price_history_map' in self.game_history and len(self.game_history['price_history_map']) > 0:
            return self.game_history['price_history_map'][-1]
    
    def get_previous_winners(self): 
        """
        Retrieves the most recent entry from the tentative winners history as a ndarray. 
        :return: The latest winners if available; otherwise, None.
        """
        if 'winner_history' in self.game_history and len(self.game_history['winner_history']) > 0:
            return self.game_history['winner_history'][-1]
    
    def get_previous_winners_map(self): 
        """
        Retrieves the most recent entry from the tentative winners history as a map. 
        :return: The latest winners if available; otherwise, None.
        """
        if 'winner_history_map' in self.game_history and len(self.game_history['winner_history_map']) > 0:
            return self.game_history['winner_history_map'][-1]
    


import json
import random
import numpy as np
import pkg_resources
import threading
from agt_server.agents.base_agents.agent import Agent
from agt_server.agents.base_agents.game_report import GameReport
from agt_server.utils import extract_first_json
import pandas as pd

class GSVM9Agent(Agent):
    def __init__(self, name=None, timestamp=None):
        super().__init__(name, timestamp)

        config_path = pkg_resources.resource_filename('agt_server', 'configs/server_configs/sa_config.json')
        with open(config_path) as cfile:
            self.config = json.load(cfile)
        self.response_time = self.config.get('response_time', 2)

        self.num_goods = None
        self.goods = None
        self.valuations = {}
        self._goods_to_index = {}
        self._index_to_goods = {}

        self._is_national_bidder = False
        self.capacity = 3 
        self.vs_str = ""

        self.timeout = False
        self.global_timeout_count = 0
        self.round_number = 0

    def timeout_handler(self):
        print(f"{self.name} has timed out")
        self.timeout = True

    def handle_permissions(self, resp):
        self.player_type = resp.get('player_type')
        gh = self.game_report.game_history

        gh.setdefault('my_bid_history', []).append(resp.get('my_bid'))
        gh.setdefault('my_utils_history', []).append(resp.get('my_util'))
        gh.setdefault('my_payment_history', []).append(resp.get('payments'))
        gh.setdefault('price_history', []).append(resp.get('prices'))
        gh.setdefault('winner_history', []).append(resp.get('winners'))

    def handle_postround_data(self, resp):
        self.global_timeout_count = resp.get('global_timeout_count', 0)
        self.handle_permissions(resp)

    def is_national_bidder(self): 
        return self._is_national_bidder
    
    def get_bids(self):
        bid_dict = {}
        for good, val in self.valuations.items():
            if val > 0:
                bid_dict[good] = val
            else:
                bid_dict[good] = 0
        return bid_dict

    def get_action(self):
        return self.get_bids()

    def print_results(self):
        gh = self.game_report.game_history
        num_rounds = len(gh.get('my_utils_history', []))
        if num_rounds == 0:
            print(f"{self.name}: No rounds have been played yet.")
            return
        
        total_util = sum(gh.get('my_utils_history', []))
        avg_util = total_util / num_rounds

        bid_list = gh.get('my_bid_history', [])
        bid_avgs = []
        for bid in bid_list:
            if bid and isinstance(bid, dict) and len(bid) > 0:
                bid_avg = sum(bid.values()) / len(bid)
                bid_avgs.append(bid_avg)
        avg_bid = sum(bid_avgs) / len(bid_avgs) if bid_avgs else 0

        # Average payment per round (assumed to be a numeric value per round)
        payment_list = gh.get('my_payment_history', [])
        avg_payment = {}
        for payment in payment_list: 
            if payment and isinstance(payment, dict):
                for good, payment_val in payment.items():
                    if payment_val is not None:
                        avg_payment[good] = avg_payment.get(good, 0) + payment_val
        

        # Count winners across rounds.
        winners_list = gh.get('winner_history', [])
        winners_freq = {}
        for winners in winners_list:
            if winners and isinstance(winners, dict):
                for good, winner in winners.items():
                    if winner is not None:
                        winners_freq[winner] = winners_freq.get(winner, 0) + 1

        # Print summary
        print(self.vs_str)
        print(f"\n===== Overall Results for {self.name} =====")
        print(f"Rounds played: {num_rounds}")
        print(f"Total Utility: {total_util:.2f}")
        print(f"Average Utility per Round: {avg_util:.2f}")
        print(f"Average Bid Value per Round: {avg_bid:.2f}")
        print("Average Payments (across goods):")
        if avg_payment: 
            for good, payment in avg_payment.items():
                avg_payment[good] = payment / num_rounds
                print(f"  {good}: {avg_payment[good]:.2f}")
        else:
            print("  No payments recorded.")
        
        print("Average #Goods Won:")
        if winners_freq:
            for winner, freq in winners_freq.items():
                print(f"  {winner}: {freq / num_rounds} Goods Won")
        else:
            print("  No Allocations recorded.")
        print("============================================\n")

    def play(self):
        """
        Main loop for interacting with the GSVM-9 (or similar) server game.
        Listens for messages and responds appropriately.
        """
        # 1. Possibly read initial handshake
        data = self.client.recv(1024).decode()
        data = extract_first_json(data)
        if data:
            resp = json.loads(data)
            if resp.get('message') == 'provide_game_name':
                print(f"We are playing {resp.get('game_name')}")
                message = {"message": "game_name_recieved"}
                self.client.send(json.dumps(message).encode())
                self.restart()

        # 2. Main loop
        while True:
            data = self.client.recv(10000).decode()
            data = extract_first_json(data)
            if not data:
                continue
            request = json.loads(data)

            msg_type = request.get('message')
            if msg_type == 'send_preround_data':

                self._is_national_bidder = request.get('player_type', "regional") == "national"
                self.capacity = 6 if self._is_national_bidder else 3
                self.vs_str = request.get('vs_str', "")
                
                self.num_goods = request.get('num_goods')
                goods_list = request.get('goods', [])
                self.goods = set(goods_list)

                # Build index mappings
                self._goods_to_index = {good: idx for idx, good in enumerate(goods_list)}
                self._index_to_goods = {idx: good for good, idx in self._goods_to_index.items()}

                # Retrieve valuations from the server
                valuations_dict = request.get('valuations', {})
                self.valuations = {}
                for good in goods_list:
                    self.valuations[good] = valuations_dict.get(good, 0.0)

                # Acknowledge
                self.client.send(json.dumps({"message": "preround_data_recieved"}).encode())

            elif msg_type == 'request_bid':
                # The server wants our bid dictionary
                self.timeout = False
                try:
                    timer = threading.Timer(self.response_time, self.timeout_handler)
                    timer.start()
                    bid = self.get_action()
                finally:
                    if self.timeout:
                        bid = None
                    timer.cancel()

                try:
                    self.client.send(json.dumps({
                        "message": "provide_bid",
                        "bid": bid,
                        "timeout": self.timeout
                    }).encode())
                except Exception as e:
                    print(f"Error sending bid: {e}")

            elif msg_type == 'prepare_next_round':
                # The server is telling us the round is over
                self.handle_postround_data(request)
                self.update()
                self.client.send(json.dumps({"message": "ready_next_round"}).encode())

            elif msg_type == 'prepare_next_game':
                # The server is telling us the game ended and a new one will begin
                self.print_results()
                self.restart()
                self.client.send(json.dumps({"message": "ready_next_game"}).encode())

            elif msg_type == 'game_end':
                # The entire game is finished
                if request.get('send_results'):
                    try:
                        df = pd.read_json(request['results'])
                        if df is not None:
                            print(df)
                    except:
                        print("Results too large. Please check with your TA.")
                else:
                    print(request.get('results'))
                self.close()
                break

            elif msg_type == 'disqualified':
                # We were disqualified
                dq_msg = request.get('disqualification_message', '')
                if dq_msg:
                    print(dq_msg)
                self.close()
                break



class Game:
    order_matters = True 
    
    def __init__(self, num_rounds, player_data, player_types, permissions_map, game_kick_timeout, game_name, invalid_move_penalty, timeout_tolerance):
        if game_name != None:
            self.game_name = game_name
        else:
            self.game_name = "Please give the Game a name"
        self.num_rounds = num_rounds
        self.player_data = player_data
        self.player_types = player_types
        self.permissions_map = permissions_map
        self.invalid_move_penalty = invalid_move_penalty
        self.timeout_tolerance = timeout_tolerance
        self.game_reports = {}
        self.kick_time = game_kick_timeout

    def run_game(self):
        raise NotImplementedError



import logging
import json
import traceback
import asyncio
import random
import numpy as np
import pandas as pd
from itertools import product
from agt_server.server.games.game import Game

class SuppressSocketSendError(logging.Filter):
    def filter(self, record):
        return "socket.send() raised exception" not in record.getMessage()

logger = logging.getLogger('asyncio')
logger.addFilter(SuppressSocketSendError())

class GSVM9Game(Game):
    def __init__(self, server_config, num_rounds=10, player_data=[], player_types=[], permissions_map={},
                 game_kick_timeout=60, game_name="GSVM-9 Auction", invalid_move_penalty=0,
                 timeout_tolerance=10):
        super().__init__(num_rounds - 1, player_data, player_types, permissions_map,
                         game_kick_timeout, game_name, invalid_move_penalty, timeout_tolerance)

        self.server_config = server_config
        self.num_goods = server_config.get("num_goods", 9)
        self.goods = server_config.get("goods", None)
        
        self.valuation_type = "complement"
        self.game_name = "GSVM-9 Auction"
        self.kth_price = server_config.get("kth_price", 1)

        self.max_valuations_national = server_config.get("max_valuations_national", None)
        self.max_valuations_regional = [
            {"A":20, "B":20, "C":40, "D":40, "E":20, "F":0,  "G":20, "H":0,  "I":0},
            {"A":0,  "B":40, "C":40, "D":20, "E":20, "F":0,  "G":0,  "H":20, "I":0},
            {"A":20, "B":20, "C":0,  "D":20, "E":20, "F":20, "G":0,  "H":0,  "I":20}
        ]
        
        self._goods_to_index = {name: idx for idx, name in enumerate(self.goods)}
        self._index_to_goods = {idx: name for name, idx in self._goods_to_index.items()}
        
        for data in self.player_data:
            self.game_reports[data['address']] = {
                "valuation_history": [],
                "bid_history": [],
                "util_history": [],
                "timeout_count": 0,
                "global_timeout_count": 0,
                "disconnected": False,
                "disqualification_message": ""
            }
        self.price_history = []
        self.winner_history = []

    async def run_game(self):
        for round in range(self.num_rounds):
            #print(f"Round: {round}")
            # --- Preround: Send valuations ---
            for i, data in enumerate(self.player_data):
                writer, reader = data['client']
                player_type = self.player_types[i]
                #print(player_type)
                # Generate valuations based on role:
                if player_type.lower() == "national":
                    valuations = {good: random.uniform(0, self.max_valuations_national[good]) for good in self.goods}
                else:
                    # Assume player_type is like "regional1", "regional2", etc.
                    try:
                        idx = int(''.join(filter(str.isdigit, player_type))) - 1
                    except:
                        idx = i - 1  # Fallback
                    valuations = {good: random.uniform(0, self.max_valuations_regional[idx][good]) for good in self.goods}
                
                self.game_reports[data['address']]["valuation_history"].append(valuations)
                message = {
                    "message": "send_preround_data",
                    "player_type": player_type,
                    "vs_str":  " VS ".join([f"{d['name']} ({self.player_types[j]})" for j, d in enumerate(self.player_data)]),
                    "num_goods": self.num_goods,
                    "goods": self.goods,
                    "valuations": valuations
                }
               
                print(" VS ".join([f"{d['name']} ({self.player_types[j]})" for j, d in enumerate(self.player_data)]))
                # print("Sending preround data to", data['name'])
                # print(message)
                if not self.game_reports[data['address']]['disconnected']:
                    try:
                        writer.write(json.dumps(message).encode())
                        await writer.drain()
                        resp = await asyncio.wait_for(reader.read(1024), timeout=self.kick_time)
                        resp = json.loads(resp)
                        assert resp['message'] == 'preround_data_recieved', f"{data['name']} did not acknowledge preround data"
                    except asyncio.TimeoutError:
                        self.game_reports[data['address']]['disqualification_message'] = f"{data['name']} Disqualified: Timed out on preround data"
                        self.game_reports[data['address']]['disconnected'] = True
                    except Exception as e:
                        stack_trace = traceback.format_exc()
                        self.game_reports[data['address']]['disqualification_message'] = f"{data['name']} Disqualified on preround data: {type(e).__name__}: {e}\n{stack_trace}"
                        self.game_reports[data['address']]['disconnected'] = True

            # --- Request Bids ---
            bids = {}
            for i, data in enumerate(self.player_data):
                writer, reader = data['client']
                if not self.game_reports[data['address']]['disconnected']:
                    message = {"message": "request_bid"}
                    #print("Sending bid request to", data['name'])
                    try:
                        writer.write(json.dumps(message).encode())
                        await writer.drain()
                        resp = await asyncio.wait_for(reader.read(1024), timeout=self.kick_time)
                        resp = json.loads(resp)
                        if resp.get('timeout', False):
                            print(f"{data['name']} timed out on bid request")
                            self.game_reports[data['address']]['bid_history'].append(None)
                            self.game_reports[data['address']]['timeout_count'] += 1
                            self.game_reports[data['address']]['global_timeout_count'] += 1
                            bids[data['name']] = None
                        else:
                            bid = resp.get('bid', None)
                            #print(bid)
                            self.game_reports[data['address']]['bid_history'].append(bid)
                            bids[data['name']] = bid
                    except asyncio.TimeoutError:
                        self.game_reports[data['address']]['disqualification_message'] = f"{data['name']} Disqualified: Timed out on bid request"
                        self.game_reports[data['address']]['disconnected'] = True
                        self.game_reports[data['address']]['bid_history'].append(None)
                        bids[data['name']] = None
                    except Exception as e:
                        stack_trace = traceback.format_exc()
                        self.game_reports[data['address']]['disqualification_message'] = f"{data['name']} Disqualified on bid request: {type(e).__name__}: {e}\n{stack_trace}"
                        self.game_reports[data['address']]['disconnected'] = True
                        self.game_reports[data['address']]['bid_history'].append(None)
                        bids[data['name']] = None
                else:
                    bids[data['name']] = None

            # --- Compute Auction Outcome ---
            #print(f"Computing auction outcome with {bids}")
            allocation, payments = self.compute_auction_result(bids)
            # print(allocation, payments)

            # --- Compute Synergy-Based Utilities ---
            for data in self.player_data:
                valuations = self.game_reports[data['address']]['valuation_history'][-1]
                # Determine which goods this player won.
                won_goods = [good for good, winner in allocation.items() if winner == data['name']]
                base_sum = sum(valuations.get(g, 0) for g in won_goods)
                n = len(won_goods)
                synergy_val = (1 + 0.2*(n-1)) * base_sum if n > 0 else 0
                total_payment = payments.get(data['name'], 0)
                round_util = synergy_val - total_payment
                self.game_reports[data['address']]['util_history'].append(round_util)

            # --- Prepare for Next Round ---
            for i, data in enumerate(self.player_data):
                player_type = self.player_types[i]
                writer, reader = data['client']
                message = {
                    "message": "prepare_next_round",
                    "player_type": player_type,
                    "permissions": self.permissions_map.get(player_type, {}),
                    "my_bid": self.game_reports[data['address']]['bid_history'][-1],
                    "my_util": self.game_reports[data['address']]['util_history'][-1],
                    "payments": payments,
                    "prices": self.price_history[-1] if self.price_history else {},
                    "winners": self.winner_history[-1] if self.winner_history else {},
                    "global_timeout_count": self.game_reports[data['address']]['global_timeout_count']
                }
                # print("Sending prepare next round to", data['name'])
                # print(message)
                try:
                    writer.write(json.dumps(message).encode())
                    await writer.drain()
                    resp = await asyncio.wait_for(reader.read(1024), timeout=self.kick_time)
                    resp = json.loads(resp)
                    assert resp['message'] == 'ready_next_round', f"{data['name']} not ready for next round"
                except asyncio.TimeoutError:
                    self.game_reports[data['address']]['disqualification_message'] = f"{data['name']} Disqualified: Timed out on prepare_next_round"
                    self.game_reports[data['address']]['disconnected'] = True
                except Exception as e:
                    stack_trace = traceback.format_exc()
                    self.game_reports[data['address']]['disqualification_message'] = f"{data['name']} error on prepare_next_round: {type(e).__name__}: {e}\n{stack_trace}"
                    self.game_reports[data['address']]['disconnected'] = True

        # --- Prepare for Next Game -- 
        for i, data in enumerate(self.player_data):
            writer, reader = data['client']
            message = {"message": "prepare_next_game"}
            # print("Sending game over to", data['name'])
            # print(message)
            if not self.game_reports[data['address']]['disconnected']:
                try:
                    writer.write(json.dumps(message).encode())
                    await writer.drain()
                    resp = await asyncio.wait_for(reader.read(1024), timeout=self.kick_time)
                    resp = json.loads(resp)
                    assert resp['message'] == 'ready_next_game', f"{data['name']} did not acknowledge game over"
                except asyncio.TimeoutError:
                    self.game_reports[data['address']]['disqualification_message'] = f"{data['name']} Disqualified: Timed out on game_over"
                    self.game_reports[data['address']]['disconnected'] = True
                except Exception as e:
                    stack_trace = traceback.format_exc()
                    self.game_reports[data['address']]['disqualification_message'] = f"{data['name']} Disqualified on game_over: {type(e).__name__}: {e}\n{stack_trace}"
                    self.game_reports[data['address']]['disconnected'] = True

        return self.game_reports

    def compute_auction_result(self, bids):
        """
        For each good, determine the highest bid among all players.
        The winner pays the kth highest bid (kth_price).
        Returns:
          allocation: dict mapping each good to the winning player's name.
          payments: dict mapping each player's name to their total payment.
        """
        allocation = {}
        payments = {data['name']: 0 for data in self.player_data}
        prices = {}
        for good in self.goods:
            bid_tuples = []
            for addr, bid_bundle in bids.items():
                if bid_bundle is not None:
                    bid_value = bid_bundle.get(good, None)
                    if bid_value is not None and bid_value > 0:
                        bid_tuples.append((bid_value, addr))
            if bid_tuples:
                sorted_bids = sorted(bid_tuples, key=lambda x: x[0], reverse=True)
                winner = sorted_bids[0][1]
                kth_index = self.kth_price - 1
                if kth_index < len(sorted_bids):
                    kth_bid = sorted_bids[kth_index][0]
                else:
                    kth_bid = sorted_bids[-1][0]
            else:
                winner = None
                kth_bid = None
            allocation[good] = winner
            prices[good] = kth_bid
            if winner is not None and kth_bid is not None:
                payments[winner] += kth_bid
        self.price_history.append(prices)
        self.winner_history.append(allocation)
        return allocation, payments

    async def run_game_final(self):
        """
        Notifies players that the game is over and they should prepare for the next game.
        """
        message = {"message": "prepare_next_game"}
        for data in self.player_data:
            if not self.game_reports[data['address']]['disconnected']:
                try:
                    writer, reader = data['client']
                    writer.write(json.dumps(message).encode())
                    await writer.drain()
                    resp = await asyncio.wait_for(reader.read(1024), timeout=self.kick_time)
                    resp = json.loads(resp)
                    assert resp['message'] == 'ready_next_game', f"{data['name']} not ready for next game"
                except asyncio.TimeoutError:
                    self.game_reports[data['address']]['disqualification_message'] = f"{data['name']} Disqualified: Timed out on prepare_next_game"
                    self.game_reports[data['address']]['disconnected'] = True
                except Exception as e:
                    stack_trace = traceback.format_exc()
                    self.game_reports[data['address']]['disqualification_message'] = f"{data['name']} error on prepare_next_game: {type(e).__name__}: {e}\n{stack_trace}"
                    self.game_reports[data['address']]['disconnected'] = True
        return self.game_reports

    def summarize_results(self):
        """
        Summarizes the game results across all rounds.
        """
        summary_data = []
        for data in self.player_data:
            addr = data['address']
            util_history = self.game_reports[addr].get('util_history', [])
            total_util = sum(util_history)
            avg_util = total_util / len(util_history) if util_history else 0
            summary_data.append([data['name'], total_util, avg_util])
        df = pd.DataFrame(summary_data, columns=["Player", "Total Utility", "Avg Utility Per Round"])
        df = df.sort_values(by="Total Utility", ascending=False)
        print(f"\nFinal GSVM-9 Auction Results after {self.num_rounds} rounds:\n", df)
        return df



from agt_server.agents.base_agents.lemonade_agent import LemonadeAgent
from agt_server.local_games.lemonade_arena import LemonadeArena
import argparse
import random
import numpy as np


class Jimbus(LemonadeAgent):
    def setup(self):
        self.opp_1_freq = np.zeros(12)
        self.opp_2_freq = np.zeros(12)
        self.alpha = 0.65


    def get_action(self):
        if(len(self.get_opp1_action_history()) == 0):
            action = random.choice(np.arange(12))
        else:
            prob1s = np.bincount(self.get_opp1_action_history(), minlength=12)/len(self.get_opp1_action_history())
            prob2s = np.bincount(self.get_opp2_action_history(), minlength=12)/len(self.get_opp2_action_history())
            scores = np.zeros([12])
            for ours in range(12):
                for other in range(12):
                    dist = min(np.abs(ours-other), np.abs(12 + ours - other), np.abs(ours - other - 12))
                    scores[ours] += prob1s[other]*dist + prob2s[other]*dist
            action = np.argmax(scores)

        return action

    def update(self):
        pass


################### SUBMISSION #####################
agent_submission = Jimbus("Jimbus")
####################################################


if __name__ == "__main__":
    #### DO NOT TOUCH THIS #####
    parser = argparse.ArgumentParser(description='My Agent')
    # parser.add_argument('agent_name', type=str, help='Name of the agent')
    parser.add_argument('--join_server', action='store_true',
                        help='Connects the agent to the server')
    parser.add_argument('--ip', type=str, default='127.0.0.1',
                        help='IP address (default: 127.0.0.1)')
    parser.add_argument('--port', type=int, default=8080,
                        help='Port number (default: 8080)')

    args = parser.parse_args()

    if args.join_server:
        agent_submission.connect(ip=args.ip, port=args.port)
    else:
        arena = LemonadeArena(
            num_rounds=1000,
            timeout=1,
            players=[
                agent_submission,
                Jimbus("Agent_1"),
                Jimbus("Agent_2"),
                Jimbus("Agent_3"),
                Jimbus("Agent_4")
            ]
        )
        arena.run()



from agt_server.agents.base_agents.lemonade_agent import LemonadeAgent
from agt_server.local_games.lemonade_arena import LemonadeArena
import argparse
import random
import numpy as np

class DelAgent(LemonadeAgent):
    def setup(self):
        self.choice = random.randint(0, 11)

    def get_action(self):
        if len(self.get_util_history()) > 10:
            utility = 0
            for i in range(10):
                utility += self.get_util_history()[-i]

            # we are doing worse than others
            if utility < 80:
                self.choice += 6
                if self.choice > 11:
                    self.choice = self.choice % 11

        return self.choice

    def update(self):
        pass

################### SUBMISSION #####################
agent_submission = DelAgent("Del")
####################################################

if __name__ == "__main__":
    #### DO NOT TOUCH THIS #####
    parser = argparse.ArgumentParser(description='My Agent')
    # parser.add_argument('agent_name', type=str, help='Name of the agent')
    parser.add_argument('--join_server', action='store_true',
                        help='Connects the agent to the server')
    parser.add_argument('--ip', type=str, default='127.0.0.1',
                        help='IP address (default: 127.0.0.1)')
    parser.add_argument('--port', type=int, default=8080,
                        help='Port number (default: 8080)')

    args = parser.parse_args()

    if args.join_server:
        agent_submission.connect(ip=args.ip, port=args.port)
    else:
        arena = LemonadeArena(
            num_rounds=1000,
            timeout=1,
            players=[
                agent_submission,
                DelAgent("Agent_1"),
                DelAgent("Agent_2"),
                DelAgent("Agent_3"),
                DelAgent("Agent_4")
            ]
        )
        arena.run()



from agt_server.agents.base_agents.sa_agent import SimultaneousAuctionAgent 
from agt_server.local_games.sa_arena import SAArena
import time
import argparse

class TruthfulAgent(SimultaneousAuctionAgent):
    def setup(self):
        pass

    def get_action(self):
        valuations = self.get_valuations() 
        bids = {}
        for good in valuations: 
            bids[good] = valuations[good]
        return bids
    
    def update(self):
        pass

################### SUBMISSION #####################
agent_submission = TruthfulAgent("Truthful Bidder")
####################################################
if __name__ == "__main__":
    #### DO NOT TOUCH THIS #####
    parser = argparse.ArgumentParser(description='My Agent')
    parser.add_argument('--join_server', action='store_true',
                        help='Connects the agent to the server')
    parser.add_argument('--ip', type=str, default='127.0.0.1',
                        help='IP address (default: 127.0.0.1)')
    parser.add_argument('--port', type=int, default=8080,
                        help='Port number (default: 8080)')

    args = parser.parse_args()

    if args.join_server:
        agent_submission.connect(ip=args.ip, port=args.port)
    else:
        arena = SAArena(
            timeout=1,
            num_goods=3,
            kth_price=2,
            players=[
                agent_submission,
                TruthfulAgent("Agent_1"),
                TruthfulAgent("Agent_2"),
                TruthfulAgent("Agent_3"),
                TruthfulAgent("Agent_4"), 
                TruthfulAgent("Agent_5"), 
                TruthfulAgent("Agent_6"), 
            ]
        )
        start = time.time()
        arena.run()
        end = time.time()
        print(f"{end - start} Seconds Elapsed")



from agt_server.agents.base_agents.lemonade_agent import LemonadeAgent
from agt_server.local_games.lemonade_arena import LemonadeArena
import argparse
import random
import numpy as np


class Kamen(LemonadeAgent):
    def setup(self):
        pass

    def get_action(self):
        try:
            if len(self.get_action_history()) < 50:
                return random.choice([2, 6, 10])
            else:
                my_acts = []

                possible_action1 = np.bincount(np.array(self.get_opp1_action_history()))
                possible_action2 = np.bincount(np.array(self.get_opp2_action_history()))

                possible_action1 = np.argpartition(possible_action1, -1)[-1:]
                possible_action2 = np.argpartition(possible_action2, -1)[-1:]

                for i, j in zip(possible_action1, possible_action2):
                    my_acts.append(self.find_action(i, j))

                return random.choice(my_acts)
        except:
            action = random.choice([10])
        return action

    def find_action(self, act1, act2):
        greatest_utils = []
        for position in range(12):
            temp = self.calculate_utility([act1, act2, position])
            greatest_utils.append(temp[-1])

        greatest_utils = np.array(greatest_utils)
        return np.random.choice(np.flatnonzero(greatest_utils == greatest_utils.max()))
    
    def update(self):
        pass

################### SUBMISSION #####################
agent_submission = Kamen("Kamen")
####################################################

if __name__ == "__main__":
    #### DO NOT TOUCH THIS #####
    parser = argparse.ArgumentParser(description='My Agent')
    # parser.add_argument('agent_name', type=str, help='Name of the agent')
    parser.add_argument('--join_server', action='store_true',
                        help='Connects the agent to the server')
    parser.add_argument('--ip', type=str, default='127.0.0.1',
                        help='IP address (default: 127.0.0.1)')
    parser.add_argument('--port', type=int, default=8080,
                        help='Port number (default: 8080)')

    args = parser.parse_args()

    if args.join_server:
        agent_submission.connect(ip=args.ip, port=args.port)
    else:
        arena = LemonadeArena(
            num_rounds=1000,
            timeout=1,
            players=[
                agent_submission,
                Kamen("Agent_1"),
                Kamen("Agent_2"),
                Kamen("Agent_3"),
                Kamen("Agent_4")
            ]
        )
        arena.run()



import logging
import json
import traceback
import asyncio
import random
import numpy as np
import pandas as pd
from itertools import product
from agt_server.server.games.game import Game
from itertools import combinations

class SuppressSocketSendError(logging.Filter):
    def filter(self, record):
        return "socket.send() raised exception" not in record.getMessage()

logger = logging.getLogger('asyncio')
logger.addFilter(SuppressSocketSendError())

class SAGame(Game):
    order_matters = False

    def __init__(self, server_config,
                 num_rounds=1000, player_data=[], player_types=[], permissions_map={},
                 game_kick_timeout=60, game_name="Simultaneous Auction", invalid_move_penalty=0,
                 timeout_tolerance=10):
        super().__init__(num_rounds, player_data, player_types, permissions_map,
                         game_kick_timeout, game_name, invalid_move_penalty, timeout_tolerance)
        self.server_config = server_config
        self.num_goods = self.server_config['num_goods']
        self.good_names = self.server_config.get('good_names', None)
        self.valuation_type = self.server_config.get('valuation_type', "additive")
        self.value_range = tuple(self.server_config.get('value_range', [0, 100]))
        self.price_history = []
        self.winner_history = []
        self.kth_price = self.server_config.get('kth_price', 1)
        self.game_name = f"{game_name} ({self.kth_price}th Price)"
        
        # Set up goods naming and mapping
        if self.good_names is not None:
            if len(self.good_names) !=self.num_goods:
                raise ValueError("Length of good_names must match num_goods")
            self._goods_to_index = {name: idx for idx, name in enumerate(self.good_names)}
            self.goods = set(self.good_names)
            self._index_to_goods = {idx: name for name, idx in self._goods_to_index.items()}
        else:
            self._goods_to_index = SAGame._name_goods(self.num_goods)
            self.goods = set(self._goods_to_index.keys())
            self._index_to_goods = {value: key for key, value in self._goods_to_index.items()}
        
        if self.valuation_type == 'additive':
            self.valuation_multiplier = 1
        elif self.valuation_type == 'complement':
            self.valuation_multiplier = 1.2
        elif self.valuation_type == 'substitute':
            self.valuation_multiplier = 0.8
        elif self.valuation_type == 'randomized':
            self.valuation_multiplier = 1.2
        
        over_bidding_parameter = 1.25
        base_value = self.value_range[1] * self.num_goods
        self.bid_upper_bound = base_value * self.valuation_multiplier ** (self.num_goods - 1) * over_bidding_parameter
        
        # Initialize game reports for each player (using address as the key)
        for data in self.player_data:
            self.game_reports[data['address']] = {
                "valuation_history": [],
                "bid_history": [],
                "util_history": [],
                "timeout_count": 0,
                "global_timeout_count": 0,
                "disconnected": False,
                "disqualification_message": ""
            }

    @staticmethod
    def _generate_sequence(alphabet='ABCDEFGHIJKLMNOPQRSTUVWXYZ'):
        yield from alphabet
        size = 2
        while True:
            for letters in product(alphabet, repeat=size):
                yield ''.join(letters)
            size += 1

    @staticmethod
    def _name_goods(size, alphabet="ABCDEFGHIJKLMNOPQRSTUVWXYZ"):
        alphabet_generator = SAGame._generate_sequence(alphabet)
        map_dict = {}
        for flat_index in range(size):
            letter = next(alphabet_generator)
            map_dict[letter] = flat_index
        return map_dict

    def compute_auction_result(self, bids):
        """
        For each good, determine the highest bid and assign the good accordingly.
        The winner pays the kth highest bid as determined by self.kth_price.
        
        Returns:
            allocation: dict mapping each good to the winning player's address.
            payments: dict mapping each player's address to their total payment.
        """
        allocation = {}
        # Initialize payments for each player by their address.
        payments = {data['name']: 0 for data in self.player_data}
        prices = {}

        for good in self.goods:
            bid_tuples = []
            # Collect all valid bids for the good.
            for addr, bid_bundle in bids.items():
                if bid_bundle is not None:
                    bid_value = bid_bundle.get(good, None)
                    if bid_value is not None and bid_value > 0:
                        bid_tuples.append((bid_value, addr))
            
            if bid_tuples:
                # Sort bids in descending order by bid value.
                sorted_bids = sorted(bid_tuples, key=lambda x: x[0], reverse=True)
                winner = sorted_bids[0][1]
                # Determine kth highest bid price.
                kth_index = self.kth_price - 1  # since list indices start at 0
                if kth_index < len(sorted_bids):
                    kth_bid = sorted_bids[kth_index][0]
                else:
                    # If there are fewer than kth bids, use the lowest bid available.
                    kth_bid = sorted_bids[-1][0]
            else:
                winner = None
                kth_bid = None
            
            allocation[good] = winner
            prices[good] = kth_bid
            if winner is not None and kth_bid is not None:
                payments[winner] += kth_bid

        self.price_history.append(prices)
        self.winner_history.append(allocation)
        return allocation, payments


    async def run_game(self):
        """
        Runs the auction game asynchronously for a set number of rounds.
        """
        print("Running auction game...")
        pairwise_adjustments = {}
        if self.valuation_type == 'randomized':
            for good1, good2 in combinations(self.goods, 2):
                pairwise_adjustments[(good1, good2)] = random.uniform(0.5, 1.5)
                    
        for round in range(self.num_rounds):
            print(f"Round {round + 1} of {self.num_rounds}")
            # --------------------
            # Preround: send valuations to each player.
            # --------------------
            for i, data in enumerate(self.player_data):
                player_type = self.player_types[i]
                writer, reader = data['client']
                # Generate random valuations for each good
                valuations = {good: random.randint(self.value_range[0], self.value_range[1]) for good in self.goods}
                self.game_reports[data['address']]["valuation_history"].append(valuations)
                pairwise_adjustments_serializable = {','.join(key): value for key, value in pairwise_adjustments.items()}
                message = {
                    "message": "send_preround_data",
                    "player_type": player_type,
                    "num_goods": self.num_goods,
                    "goods": list(self.goods),
                    "valuations": valuations,
                    "pairwise_adjustments": pairwise_adjustments_serializable,
                }
                
                if not self.game_reports[data['address']]['disconnected']:
                    try:
                        writer.write(json.dumps(message).encode())
                        await writer.drain()
                        resp = await asyncio.wait_for(reader.read(1024), timeout=self.kick_time)
                        resp = json.loads(resp)
                        assert resp['message'] == 'preround_data_recieved', (
                            f"{data['name']} did not acknowledge preround data"
                        )
                    except asyncio.TimeoutError:
                        #print(f"{data['name']} Disqualified: Timed out on preround data")
                        self.game_reports[data['address']]['disqualification_message'] = (
                            f"{data['name']} Disqualified: Timed out on preround data"
                        )
                        self.game_reports[data['address']]['disconnected'] = True
                    except Exception as e:
                        stack_trace = traceback.format_exc()
                        #print(f"{data['name']} Disqualified on preround data: {type(e).__name__}: {e}\n{stack_trace}")
                        self.game_reports[data['address']]['disqualification_message'] = (
                            f"{data['name']} Disqualified on preround data: {type(e).__name__}: {e}\n{stack_trace}"
                        )
                        self.game_reports[data['address']]['disconnected'] = True

            # --------------------
            # Request bids from each player.
            # --------------------
            bids = {}
            for i, data in enumerate(self.player_data):
                writer, reader = data['client']
                if not self.game_reports[data['address']]['disconnected']:
                    message = {"message": "request_bid"}
                    try:
                        writer.write(json.dumps(message).encode())
                        await writer.drain()
                        resp = await asyncio.wait_for(reader.read(1024), timeout=self.kick_time)
                        resp = json.loads(resp)
                        if resp.get('timeout', False):
                            #print(f"{data['name']} Disqualified: Timed out on bid")
                            self.game_reports[data['address']]['bid_history'].append(None)
                            self.game_reports[data['address']]['timeout_count'] += 1
                            self.game_reports[data['address']]['global_timeout_count'] += 1
                            bids[data['name']] = None
                        else:
                            # Expecting a bid dictionary
                            bid = resp.get('bid', None)
                            self.game_reports[data['address']]['bid_history'].append(bid)
                            bids[data['name']] = bid
                            #print(f"Received bid from {data['name']}: {bid}")
                    except asyncio.TimeoutError:
                        self.game_reports[data['address']]['disqualification_message'] = (
                            f"{data['name']} Disqualified: Timed out on bid request"
                        )
                        #print(f"{data['name']} Disqualified: Timed out on bid request")
                        self.game_reports[data['address']]['disconnected'] = True
                        self.game_reports[data['address']]['bid_history'].append(None)
                        bids[data['name']] = None
                    except Exception as e:
                        stack_trace = traceback.format_exc()
                        self.game_reports[data['address']]['disqualification_message'] = (
                            f"{data['name']} Disqualified on bid request: {type(e).__name__}: {e}\n{stack_trace}"
                        )
                        #print(f"{data['name']} Disqualified on bid request: {type(e).__name__}: {e}\n{stack_trace}")
                        self.game_reports[data['address']]['disconnected'] = True
                        self.game_reports[data['address']]['bid_history'].append(None)
                        bids[data['name']] = None
                else:
                    #print("Warning: Player disconnected, skipping bid request.")
                    bids[data['name']] = None

            # --------------------
            # Compute auction outcome
            # --------------------
            allocation, payments = self.compute_auction_result(bids)


            # --------------------
            # Calculate round utilities for each player.
            # --------------------
            for data in self.player_data:
                valuations = self.game_reports[data['address']]['valuation_history'][-1]
                # Determine which goods the player won this round
                won_goods = [good for good, winner in allocation.items() if winner == data['name']]
                base_sum = sum(valuations.get(good, 0) for good in won_goods)
                total_payment = payments.get(data['name'], 0)
                n = len(won_goods)
                vt = self.valuation_type
                if vt == 'additive':
                    round_val = base_sum
                elif vt == 'complement':
                    round_val = base_sum * (1 + 0.05 * (n - 1)) if n > 0 else 0
                elif vt == 'substitute':
                    round_val = base_sum * (1 - 0.05 * (n - 1)) if n > 0 else 0
                elif vt == 'randomized':
                    new_valuations = valuations.copy()
                    for good1, good2 in combinations(won_goods, 2):
                        multiplier = pairwise_adjustments.get((good1, good2), 1.0)
                        new_valuations[good1] *= multiplier
                        new_valuations[good2] *= multiplier
                    
                    round_val = sum(new_valuations.get(good, 0) for good in won_goods)
                else:
                    round_val = base_sum
                round_util = round_val - total_payment
                self.game_reports[data['address']]['util_history'].append(round_util)
            
            # --------------------
            # Prepare for next round: send update message to players.
            # --------------------
            for i, data in enumerate(self.player_data):
                player_type = self.player_types[i]
                writer, reader = data['client']
                opp_bid_history = [bids.get(other['name'], None) 
                       for other in self.player_data 
                       if other['address'] != data['address']]
                 
                message = {
                    "message": "prepare_next_round",
                    "player_type": player_type,
                    "permissions": self.permissions_map.get(player_type, {}),
                    "my_bid": self.game_reports[data['address']]['bid_history'][-1],
                    "my_util": self.game_reports[data['address']]['util_history'][-1],
                    "payments": payments,
                    "prices": self.price_history[-1],
                    "winners": self.winner_history[-1],
                    "opp_bid_history": opp_bid_history,
                    "bid_upper_bound": self.bid_upper_bound,
                    "global_timeout_count": self.game_reports[data['address']]['global_timeout_count']
                }
                try:
                    writer.write(json.dumps(message).encode())
                    await writer.drain()
                    resp = await asyncio.wait_for(reader.read(1024), timeout=self.kick_time)
                    resp = json.loads(resp)
                    assert resp['message'] == 'ready_next_round', f"{data['name']} not ready for next round"
                except asyncio.TimeoutError:
                    self.game_reports[data['address']]['disqualification_message'] = (
                        f"{data['name']} Disqualified: Timed out on prepare_next_round"
                    )
                    #print(f"{data['name']} Disqualified: Timed out on prepare_next_round")
                    self.game_reports[data['address']]['disconnected'] = True
                except Exception as e:
                    stack_trace = traceback.format_exc()
                    self.game_reports[data['address']]['disqualification_message'] = (
                        f"{data['name']} error on prepare_next_round: {type(e).__name__}: {e}\n{stack_trace}"
                    )
                    #print(f"{data['name']} error on prepare_next_round: {type(e).__name__}: {e}\n{stack_trace}")
                    self.game_reports[data['address']]['disconnected'] = True

        return self.game_reports

    def summarize_results(self):
        summary_data = []
        for data in self.player_data:
            addr = data['address']
            util_history = self.game_reports[addr].get('util_history', [])
            total_util = sum(util_history)
            avg_util = total_util / len(util_history) if util_history else 0
            summary_data.append([data['name'], total_util, avg_util])
        df = pd.DataFrame(summary_data, columns=["Player", "Total Utility", "Avg Utility Per Round"])
        df = df.sort_values(by="Total Utility", ascending=False)
        print(f"\nFinal Auction Results after {self.num_rounds} rounds:\n", df)
        return df

    def print_results(self, data):
        total_util = sum(self.game_reports[data['address']].get('util_history', []))
        print(f"{data['name']} got a total auction utility of {total_util}")
        return total_util



from agt_server.agents.base_agents.agent import Agent
from agt_server.agents.base_agents.game_report import GameReport
from agt_server.utils import extract_first_json
import json
import random
import numpy as np 
import pkg_resources
import threading
import pandas as pd
from itertools import combinations

class SimultaneousAuctionAgent(Agent):
    def __init__(self, name=None, timestamp=None):
        super().__init__(name, timestamp)
        config_path = pkg_resources.resource_filename('agt_server', 'configs/server_configs/sa_config.json')
        with open(config_path) as cfile:
            self.config = json.load(cfile)
        self.response_time = self.config['response_time']

        # To be set by the game
        self.num_goods = None
        self.valuation_type = None
        self.valuations = None
        self._goods_to_index = None
        self.goods = None
        self._index_to_goods = None
        self.round = 0 
        
        self.pairwise_adjustments = None 

    def timeout_handler(self):
        print(f"{self.name} has timed out")
        self.timeout = True

    def handle_permissions(self, resp):
        self.player_type = resp['player_type']
        self.game_report.game_history['my_bid_history'].append(resp['my_bid'])
        self.game_report.game_history['my_utils_history'].append(resp['my_util'])
        self.game_report.game_history['my_payment_history'].append(resp['payments'])
        self.game_report.game_history['price_history'].append(resp['prices'])
        self.game_report.game_history['winner_history'].append(resp['winners'])
        opp_bids = [bid for bid in resp['opp_bid_history'] if bid is not None]
        self.game_report.game_history['opp_bid_history'].append(opp_bids)
    
    def handle_postround_data(self, resp):
        self.global_timeout_count = resp['global_timeout_count']
        self.handle_permissions(resp)

    def get_action(self):
        # This should return the bid dictionary.
        return self.get_bids()
     
    def get_goods(self): 
        """
        Get the set of goods names available in the auction.
        """
        return self.goods 

    def get_num_goods(self): 
        """
        Get the total number of goods available.
        """
        return self.num_goods
    
    def get_goods_to_index(self): 
        """
        Get the mapping from goods names to their index.
        """
        return self._goods_to_index
    
    def calculate_valuation(self, goods):
        """
        Calculate the valuation for a given set of goods.
        """
        # Convert goods to indices if necessary
        
        if all(isinstance(good, str) for good in goods):
            goods = [self._goods_to_index[good] for good in goods]
        
        base_sum = np.sum([self.valuations[good] for good in goods])
        vt = self.valuation_type
        n = len(goods)

        if vt == 'additive':
            tot_val = base_sum
        elif vt == 'complement':
            tot_val = base_sum * (1 + 0.05 * (n - 1)) if n > 0 else 0
        elif vt == 'substitute':
            tot_val = base_sum * (1 - 0.05 * (n - 1)) if n > 0 else 0
        elif vt == 'randomized':
            new_valuations = self.valuations.copy()
            for good1, good2 in combinations(goods, 2):
                multiplier = self.pairwise_adjustments.get((good1, good2), 1.0)
                new_valuations[good1] *= multiplier
                new_valuations[good2] *= multiplier
            
            tot_val = sum(new_valuations[good] for good in goods)
        else:
            tot_val = base_sum
            
        return tot_val
    
    def get_valuation_as_array(self): 
        """
        Retrieves the agent's valuation as a numpy array.
        """
        return self.valuations
    
    def get_valuation(self, good): 
        """
        Retrieves the valuation for a specific good.
        """
        return self.valuations[self._goods_to_index[good]]
    
    def get_valuations(self, bundle=None): 
        """
        Retrieves the valuations for a set of goods.
        """
        if bundle is None: 
            bundle = self.goods
        return {good: self.valuations[self._goods_to_index[good]] for good in bundle}
    
    def get_game_report(self): 
        """
        Retrieves the game report.
        """
        return self.game_report
    
    def get_valuation_history(self):
        """
        Retrieves the valuation history.
        """
        return self.game_report.get_valuation_history()

    def get_util_history(self):
        """
        Retrieves the utility history.
        """
        return self.game_report.get_util_history()

    def get_bid_history(self): 
        """
        Retrieves the history of bids made.
        """
        return self.game_report.get_bid_history()
    
    def get_payment_history(self):
        """
        Retrieves the history of payments made.
        """
        return self.game_report.get_payment_history()
    
    def get_price_history(self):
        """
        Retrieves the history of prices.
        """
        return self.game_report.get_price_history()
    
    def get_winner_history(self): 
        """
        Retrieves the history of winners.
        """
        return self.game_report.get_winner_history()
    
    def print_results(self):
        """
        Prints a summary of the current game results based on the game history.
        This includes details for the most recent round as well as cumulative statistics.
        """
        gh = self.game_report.game_history  # shorthand
        
        # Retrieve the most recent round data (if any)
        last_bid = gh.get('my_bid_history', [])[-1] if gh.get('my_bid_history') else None
        last_winners = gh.get('winner_history', [])[-1] if gh.get('winner_history') else None
        last_prices = gh.get('price_history', [])[-1] if gh.get('price_history') else None

        # Print the results for the most recent round
        if self.round > 0:
            print(f"Round {self.round} Results:")
            print("-----------------------------------------")
            print(f"Last Bid: {last_bid}")
            print(f"Winners: {last_winners}")
            print(f"Prices: {last_prices}")
            print("-----------------------------------------\n")
            
        self.round += 1
    def play(self):
        """
        Main loop for interacting with the Simultaneous Auction server.
        Listens for messages and responds appropriately:
          - Acknowledges preround data and updates auction parameters.
          - Responds to bid requests by sending its bid (via get_bids).
          - Handles round and game transitions.
          - Processes disqualification messages.
        """
        # Initial handshake (if any)
        data = self.client.recv(1024).decode()
        data = extract_first_json(data)
        if data:
            resp = json.loads(data)
            if resp.get('message') == 'provide_game_name':
                print(f"We are playing {resp.get('game_name')}")
                message = {"message": "game_name_recieved"}
                self.client.send(json.dumps(message).encode())
                self.restart()
        
        while True:
            data = self.client.recv(10000).decode()
            data = extract_first_json(data)
            if data:
                request = json.loads(data)
                if request.get('message') == 'send_preround_data':
                    # Update auction parameters from the server
                    self.player_type = request.get('player_type')
                    self.num_goods = request.get('num_goods')
                    goods_list = request.get('goods')
                    self.goods = set(goods_list)
                    valuations_dict = request.get('valuations')
                    raw_adjustments = request.get('pairwise_adjustments', {})
                    self.pairwise_adjustments = {tuple(key.split(',')): value for key, value in raw_adjustments.items()}
                    
                    # Create goods mapping from the provided goods list
                    self._goods_to_index = {good: idx for idx, good in enumerate(goods_list)}
                    self._index_to_goods = {idx: good for good, idx in self._goods_to_index.items()}
                    self.valuations = np.zeros(self.num_goods)
                    for good, value in valuations_dict.items():
                        idx = self._goods_to_index[good]
                        self.valuations[idx] = value
                    message = {"message": "preround_data_recieved"}
                    self.client.send(json.dumps(message).encode())
                    continue
                elif request.get('message') == 'request_bid':
                    self.timeout = False
                    try:
                        timer = threading.Timer(self.response_time, self.timeout_handler)
                        timer.start()
                        bid = self.get_action()  # This should return your bid dictionary
                    finally:
                        if self.timeout:
                            bid = None
                        timer.cancel()
                    try:
                        message = {
                            "message": "provide_bid",
                            "bid": bid,
                            "timeout": self.timeout
                        }
                        self.client.send(json.dumps(message).encode())
                    except Exception as e:
                        print("Error sending bid:", e)
                elif request.get('message') == 'prepare_next_round':
                    self.print_results()
                    self.handle_postround_data(request)
                    self.update()
                    message = {"message": "ready_next_round"}
                    self.client.send(json.dumps(message).encode())
                elif request['message'] == 'game_end':
                    if request['send_results']:
                        try: 
                            df = pd.read_json(request['results'])
                            if df is not None:
                                print(df)
                        except: 
                            print("Results too large :(. Please check in with your Lab TA for the results")
                    else:
                        print(request['results'])
                    self.close()
                    break
                elif request.get('message') == 'disqualified':
                    if request.get('disqualification_message'):
                        print(request.get('disqualification_message'))
                    self.close()
                    break



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()
with open('requirements.txt') as f:
    required = f.read().splitlines()

setup(
    name='agt_server', 
    version='1.9.15',
    author='John Wu', 
    author_email='john_w_wu@brown.edu', 
    description='The AGT Server is a python platform designed to run and implement game environments that autonomous agents can connect to and compete in.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/brown-agt/agt-server-remastered',  
    project_urls={
        "Bug Tracker": "https://github.com/brown-agt/agt-server-remastered/issues",  
    },
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    package_dir={"": "src"},
    package_data={'agt_server': ['configs/server_configs/*.json', 'configs/handin_configs/*.json']},
    packages=find_packages(where="src"),
    python_requires='>=3.10',
    install_requires=required,
)


import threading
import traceback
from datetime import datetime
import os
import json

class LocalArena:
    def __init__(self, num_rounds, players, timeout, handin, logging_path, save_path):
        self.num_rounds = num_rounds
        self.players = players
        self.timeout = timeout
        self.handin_mode = handin
        self.timeout_tolerance = 5
        self.game_reports = {}

        # Initialize logging and save paths
        self.logging_path = logging_path
        self.save_path = save_path

        # Create a timestamped log file if logging_path is provided
        if self.logging_path:
            timestamp = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
            self.logging_path = os.path.join(logging_path, f"{timestamp}_log.txt")

    def run_func_w_time(self, func, timeout, name, alt_ret=None):
        """
        Runs a function with a timeout, capturing exceptions and handling timeouts.
        """
        ret = None
        completed = threading.Event()

        def target_wrapper():
            nonlocal ret
            try:
                ret = func()
            except Exception as e:
                self._log_or_print_exception(e, name)
            finally:
                completed.set()

        thread = threading.Thread(target=target_wrapper)
        thread.start()
        thread.join(timeout)

        # Handle timeout
        if not completed.is_set():
            self._handle_timeout(name)
            return alt_ret 

        return ret

    def _log_or_print_exception(self, exception, name):
        """
        Logs or prints exceptions during function execution.
        """
        stack_trace = traceback.format_exc()
        message = f"Exception in {name}: {exception}\n{stack_trace}"
        self._log_or_print(message)

        # Mark the player as disconnected if in handin mode
        if self.handin_mode and name in self.game_reports:
            self.game_reports[name]["disconnected"] = True

    def _handle_timeout(self, name):
        """
        Handles timeout events for a player.
        """
        if name in self.game_reports:
            self.game_reports[name].setdefault("timeout_count", 0)
            self.game_reports[name].setdefault("global_timeout_count", 0)
            self.game_reports[name]["timeout_count"] += 1
            self.game_reports[name]["global_timeout_count"] += 1

        message = f"{name} timed out after {self.timeout} seconds."
        self._log_or_print(message)

    def run_game(self):
        """
        Placeholder for the game logic. Must be implemented in subclasses.
        """
        raise NotImplementedError

    def _log_or_print(self, message):
        """
        Helper function to log messages to a file if logging_path is set, otherwise print them.
        """
        if self.logging_path:
            with open(self.logging_path, "a") as log_file:
                log_file.write(message + "\n")
        else:
            print(message)



from agt_server.agents.base_agents.chicken_agent import ChickenAgent
from agt_server.local_games.chicken_arena import ChickenArena
import argparse
import numpy as np


class ExponentialAgent(ChickenAgent):
    def __init__(self, name):
        super(ExponentialAgent, self).__init__(name)
        self.setup()

    def setup(self):
        self.SWERVE, self.CONTINUE = 0, 1
        self.actions = [self.SWERVE, self.CONTINUE]
        self.my_utils = np.zeros(len(self.actions))
        self.counts = [0, 0]

        # NOTE: Changing this will only change your agent's perception of the utility and
        #       will not change the actual utility used in the game
        self.utility = [[0, -1], [1, -5]]

    @staticmethod
    def softmax(x):
        # Shifting values to avoid nan issues (due to underflow)
        shifted_x = x - np.max(x)
        exp_values = np.exp(shifted_x)
        return exp_values/np.sum(exp_values)

    def get_action(self):
        move_p = self.calc_move_probs()
        my_move = np.random.choice(self.actions, p=move_p)
        return my_move

    def update(self):
        """
        HINT: Update your move history and utility history to help find your best move in calc_move_probs
        """
        my_last_action = self.get_last_action()
        my_last_utility = self.get_last_util()
        self.my_utils[my_last_action] += my_last_utility
        self.counts[my_last_action] += 1

    def calc_move_probs(self):
        """
         Uses your historical average rewards to generate a probability distribution over your next move using
         the Exponential Weights strategy

         Please return an array of [P(COOPERATE), P(STUBBORN)]
        """
        # TODO Calculate the average reward for each action over time and return the softmax of it
        average_util = np.zeros(len(self.actions))
        for i, _ in enumerate(self.actions):
            average_util[i] = self.my_utils[i]
            if self.counts[i] != 0:
                average_util[i] = average_util[i] / self.counts[i]
        return self.softmax(average_util)

################### SUBMISSION #####################
agent_submission = ExponentialAgent("Exponential Agent")
####################################################

if __name__ == "__main__":
    #### DO NOT TOUCH THIS #####
    parser = argparse.ArgumentParser(description='My Agent')
    parser.add_argument('agent_name', type=str, help='Name of the agent')
    parser.add_argument('--join_server', action='store_true',
                        help='Connects the agent to the server')
    parser.add_argument('--ip', type=str, default='127.0.0.1',
                        help='IP address (default: 127.0.0.1)')
    parser.add_argument('--port', type=int, default=8080,
                        help='Port number (default: 8080)')

    args = parser.parse_args()

    agent = ExponentialAgent(args.agent_name)
    if args.join_server:
        agent.connect(ip=args.ip, port=args.port)
    else:
        arena = ChickenArena(
            num_rounds=1000,
            timeout=1,
            players=[
                agent,
                ExponentialAgent("Agent_1"),
                ExponentialAgent("Agent_2"),
                ExponentialAgent("Agent_3"),
                ExponentialAgent("Agent_4")
            ]
        )
        arena.run()



from agt_server.agents.base_agents.lemonade_agent import LemonadeAgent
from agt_server.local_games.lemonade_arena import LemonadeArena
import argparse
import random
import numpy as np


class HiBot(LemonadeAgent):
    def setup(self):
        self.opp_1_freq = np.zeros(12)
        self.opp_2_freq = np.zeros(12)
        self.alpha = 0.65


    def get_action(self):
        try:
            last_opp_1 = self.get_opp1_action_history()[-1]
            last_opp_2 = self.get_opp2_action_history()[-1]

            self.opp_1_freq *= self.alpha
            self.opp_2_freq *= self.alpha

            self.opp_1_freq[last_opp_1] += 1
            self.opp_2_freq[last_opp_2] += 1

            player_1_prob = np.exp(self.opp_1_freq) / np.sum(np.exp(self.opp_1_freq))
            player_2_prob = np.exp(self.opp_2_freq) / np.sum(np.exp(self.opp_2_freq))

            opp_1_exp = np.random.choice(np.arange(12), p=player_1_prob)
            opp_2_exp = np.random.choice(np.arange(12), p=player_2_prob)

            # opp_1_exp = np.argmax(self.opp_1_freq)
            # opp_2_exp = np.argmax(self.opp_2_freq)

            # rel = (opp_1_exp - opp_2_exp ) % 12

            if opp_1_exp >= opp_2_exp:
                greater = opp_1_exp
                rel = opp_1_exp - opp_2_exp
            else:
                greater = opp_2_exp
                rel = opp_2_exp - opp_1_exp

            if rel <= 6:
                action = (greater + 1) % 12
            else:
                action = (greater - 1) % 12

        except:
            action = np.random.choice(np.arange(12))
        return action

    def update(self):
        pass

################### SUBMISSION #####################
agent_submission = HiBot("Hi")
####################################################

if __name__ == "__main__":
    #### DO NOT TOUCH THIS #####
    parser = argparse.ArgumentParser(description='My Agent')
    # parser.add_argument('agent_name', type=str, help='Name of the agent')
    parser.add_argument('--join_server', action='store_true',
                        help='Connects the agent to the server')
    parser.add_argument('--ip', type=str, default='127.0.0.1',
                        help='IP address (default: 127.0.0.1)')
    parser.add_argument('--port', type=int, default=8080,
                        help='Port number (default: 8080)')

    args = parser.parse_args()

    if args.join_server:
        agent_submission.connect(ip=args.ip, port=args.port)
    else:
        arena = LemonadeArena(
            num_rounds=1000,
            timeout=1,
            players=[
                agent_submission,
                HiBot("Agent_1"),
                HiBot("Agent_2"),
                HiBot("Agent_3"),
                HiBot("Agent_4")
            ]
        )
        arena.run()



from agt_server.agents.base_agents.agent import Agent
from agt_server.utils import extract_first_json
import json
import pandas as pd
import threading
import pkg_resources


class LemonadeAgent(Agent):
    def __init__(self, name=None, timestamp=None):
        super().__init__(name, timestamp)
        self.valid_actions = list(range(12))
        config_path = pkg_resources.resource_filename('agt_server', 'configs/server_configs/lemonade_config.json')
        with open(config_path) as cfile:
            server_config = json.load(cfile)
        self.response_time = server_config['response_time']

    def timeout_handler(self):
        print(f"{self.name} has timed out")
        self.timeout = True

    def handle_permissions(self, resp):
        self.player_type = resp['player_type']
        if 'all' in resp['permissions']:
            self.game_report.game_history['my_action_history'].append(
                resp['my_action'])
            self.game_report.game_history['my_utils_history'].append(
                resp['my_utils'])
            self.game_report.game_history['opp1_action_history'].append(
                resp['opp1_action'])
            self.game_report.game_history['opp1_utils_history'].append(
                resp['opp1_utils'])
            self.game_report.game_history['opp2_action_history'].append(
                resp['opp2_action'])
            self.game_report.game_history['opp2_utils_history'].append(
                resp['opp2_utils'])
        else:
            for perm in resp['permissions']:
                self.game_report.game_history[f'{perm}_history'].append(
                    resp[perm])

    def handle_postround_data(self, resp):
        self.global_timeout_count = resp['global_timeout_count']
        self.curr_opps = resp['opp_names']
        self.handle_permissions(resp)

    def play(self):
        data = self.client.recv(1024).decode()
        data = extract_first_json(data)
        
        if data:
            resp = json.loads(data)
            if resp['message'] == 'provide_game_name':
                print(f"We are playing {resp['game_name']}")
                message = {
                    "message": "game_name_recieved",
                }
                self.client.send(json.dumps(message).encode())
                self.restart()
        while True:
            data = self.client.recv(10000).decode()
            data = extract_first_json(data)
            
            if data:
                request = json.loads(data)
                if request['message'] == 'send_preround_data':
                    self.player_type = request['player_type']
                    message = {"message": "preround_data_recieved"}
                    self.client.send(json.dumps(message).encode())
                    continue
                elif request['message'] == 'request_action':
                    self.timeout = False
                    try:
                        timer = threading.Timer(self.response_time, self.timeout_handler)
                        timer.start()
                        action = self.get_action()
                    finally:
                        if self.timeout: 
                            action = -1
                        timer.cancel()
                    
                    try:
                        action = int(action)
                        message = {
                            "message": "provide_action",
                            "action": action, 
                            "timeout": self.timeout
                        }
                        json_m = json.dumps(message).encode()
                        self.client.send(json_m)
                    except:
                        print("Warning: Get Action must return an Integer")
                        message = {
                            "message": "provide_action",
                            "action": -1, 
                            "timeout": self.timeout
                        }
                        json_m = json.dumps(message).encode()
                        self.client.send(json_m)
                elif request['message'] == 'game_end':
                    if request['send_results']:
                        try: 
                            df = pd.read_json(request['results'])
                            if df is not None:
                                print(df)
                        except: 
                            print("Results too large :(. Please check in with your Lab TA for the results")
                    else:
                        print(request['results'])
                    self.close()
                    break

            data = self.client.recv(10000).decode()
            data = extract_first_json(data)
            
            if data:
                resp = json.loads(data)
                if resp['message'] == 'prepare_next_game':
                    self.print_results()
                    self.restart()
                    message = {"message": "ready_next_game"}
                    self.client.send(json.dumps(message).encode())
                elif resp['message'] == 'prepare_next_round':
                    self.handle_postround_data(resp)
                    self.update()
                    message = {"message": "ready_next_round"}
                    self.client.send(json.dumps(message).encode())
                elif resp['message'] == 'disqualified':
                    if resp['disqualification_message']:
                        print(resp['disqualification_message'])
                    self.close()
                    break

    def print_results(self):
        action_counts = [0 for _ in range(len(self.valid_actions) + 1)]
        for action in self.game_report.game_history['my_action_history']:
            if action in self.valid_actions:
                action_counts[action] += 1
            else:
                action_counts[len(self.valid_actions)] += 1
        print_smt = f"{self.name} set up their Lemonade Stand at"
        for i in range(len(self.valid_actions) - 1):
            print_smt += f" Location {i} {action_counts[i]} times,"
        print_smt += f" and Location {len(self.valid_actions) - 1} {action_counts[len(self.valid_actions) - 1]} times."
        print(f"Game {self.game_num}:")
        if self.curr_opps: 
            and_str = ''
            if len(self.curr_opps) > 1: 
                and_str += ' and '
            print(f"I am currently playing against {', '.join(self.curr_opps[:-1]) + and_str + self.curr_opps[-1]}")
        print(print_smt)
        if action_counts[len(self.valid_actions)] > 0:
            print(
                f"{self.name} submitted {action_counts[len(self.valid_actions)]} invalid moves")

        if self.global_timeout_count > 0:
            print(f"{self.name} timed out {self.global_timeout_count} times")

        total_util = sum(self.game_report.game_history['my_utils_history'])

        avg_util = total_util / \
            len(self.game_report.game_history['my_utils_history'])
        print(
            f"{self.name} got a total utility of {total_util} and a average utility of {avg_util}")
        self.game_num += 1

    def get_game_report(self): 
        return self.game_report
    
    def get_action_history(self):
        return self.game_report.get_action_history()

    def get_util_history(self):
        return self.game_report.get_util_history()

    def get_opp1_action_history(self):
        return self.game_report.get_opp1_action_history()

    def get_opp1_util_history(self):
        return self.game_report.get_opp1_util_history()

    def get_opp2_action_history(self):
        return self.game_report.get_opp2_action_history()

    def get_opp2_util_history(self):
        return self.game_report.get_opp2_util_history()

    def get_last_action(self):
        return self.game_report.get_last_action()

    def get_last_util(self):
        return self.game_report.get_previous_util()

    def get_opp1_last_action(self):
        return self.game_report.get_opp1_last_action()

    def get_opp1_last_util(self):
        return self.game_report.get_opp1_last_util()

    def get_opp2_last_action(self):
        return self.game_report.get_opp2_last_action()

    def get_opp2_last_util(self):
        return self.game_report.get_opp2_last_util()

    def calculate_utility(self, p1_action, p2_action, p3_action):
        utils = [0, 0, 0]
        if p1_action not in self.valid_actions and p2_action not in self.valid_actions and p3_action not in self.valid_actions:
            utils = [0, 0, 0]
        elif p1_action not in self.valid_actions and p2_action not in self.valid_actions:
            utils = [0, 0, 24]
        elif p1_action not in self.valid_actions and p3_action not in self.valid_actions:
            utils = [0, 24, 0]
        elif p2_action not in self.valid_actions and p3_action not in self.valid_actions:
            utils = [24, 0, 0]
        elif p1_action not in self.valid_actions:
            utils = [0, 12, 12]
        elif p2_action not in self.valid_actions:
            utils = [12, 0, 12]
        elif p3_action not in self.valid_actions:
            utils = [12, 12, 0]
        elif p1_action == p2_action and p2_action == p3_action:
            utils = [8, 8, 8]
        elif p1_action == p2_action:
            utils = [6, 6, 12]
        elif p1_action == p3_action:
            utils = [6, 12, 6]
        elif p2_action == p3_action:
            utils = [12, 6, 6]
        else:
            actions = [p1_action, p2_action, p3_action]
            sorted_actions = sorted(actions)
            index_map = {action: index for index,
                         action in enumerate(actions)}
            sorted_indices = [index_map[action]
                              for action in sorted_actions]
            u1 = sorted_actions[1] - sorted_actions[0]
            u2 = sorted_actions[2] - sorted_actions[1]
            u3 = 12 + sorted_actions[0] - sorted_actions[2]
            utils[sorted_indices[0]] = u1 + u3
            utils[sorted_indices[1]] = u1 + u2
            utils[sorted_indices[2]] = u2 + u3
        return utils
